import functools
import logging

from eth_abi.exceptions import DecodingError
from eth_typing import ChecksumAddress
from web3.exceptions import Web3Exception

from gnosis.eth import EthereumClient

from .exceptions import CannotGetPriceFromOracle

logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=10_000)
def get_decimals(
    token_address: ChecksumAddress, ethereum_client: EthereumClient
) -> int:
    """
    Auxiliary function so RPC call to get `token decimals` is cached and
    can be reused for every Oracle, instead of having a cache per Oracle.

    :param token_address:
    :param ethereum_client:
    :return: Decimals for a token
    :raises CannotGetPriceFromOracle: If there's a problem with the query
    """
    try:
        return ethereum_client.erc20.get_decimals(token_address)
    except (
        Web3Exception,
        DecodingError,
        ValueError,
    ) as e:
        message = f"Cannot get decimals for token={token_address}"
        logger.debug(message)
        raise CannotGetPriceFromOracle(message) from e



import binascii
from typing import Any, Optional

from django import forms
from django.core import exceptions
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

from hexbytes import HexBytes

from gnosis.eth.utils import fast_is_checksum_address


class EthereumAddressFieldForm(forms.CharField):
    default_error_messages = {
        "invalid": _("Enter a valid checksummed Ethereum Address."),
    }

    def prepare_value(self, value):
        return value

    def to_python(self, value):
        value = super().to_python(value)
        if value in self.empty_values:
            return None
        elif not fast_is_checksum_address(value):
            raise ValidationError(self.error_messages["invalid"], code="invalid")
        return value


class HexFieldForm(forms.CharField):
    default_error_messages = {
        "invalid": _("Enter a valid hexadecimal."),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.empty_value = None

    def prepare_value(self, value: memoryview) -> str:
        if value:
            return "0x" + bytes(value).hex()
        else:
            return ""

    def to_python(self, value: Optional[Any]) -> Optional[HexBytes]:
        if value in self.empty_values:
            return self.empty_value
        try:
            if isinstance(value, str):
                value = value.strip()
            return HexBytes(value)
        except (binascii.Error, TypeError, ValueError):
            raise exceptions.ValidationError(
                self.error_messages["invalid"],
                code="invalid",
                params={"value": value},
            )


class Keccak256FieldForm(HexFieldForm):
    default_error_messages = {
        "invalid": _('"%(value)s" is not a valid keccak256 hash.'),
        "length": _('"%(value)s" keccak256 hash should be 32 bytes.'),
    }

    def prepare_value(self, value: str) -> str:
        # Keccak field already returns a hex str
        return value

    def to_python(self, value: Optional[Any]) -> HexBytes:
        python_value: Optional[HexBytes] = super().to_python(value)
        if python_value and len(python_value) != 32:
            raise ValidationError(
                self.error_messages["length"],
                code="length",
                params={"value": python_value.hex()},
            )
        return python_value



import json
from typing import Any, Dict, Optional
from urllib.parse import urljoin

import requests
from eth_typing import ChecksumAddress

from .. import EthereumNetwork
from .contract_metadata import ContractMetadata


class BlockscoutClientException(Exception):
    pass


class BlockScoutConfigurationProblem(BlockscoutClientException):
    pass


class BlockscoutClient:
    NETWORK_WITH_URL = {
        EthereumNetwork.GNOSIS: "https://gnosis.blockscout.com/api/v1/graphql",
        EthereumNetwork.ENERGY_WEB_CHAIN: "https://explorer.energyweb.org/graphiql",
        EthereumNetwork.ENERGY_WEB_VOLTA_TESTNET: "https://volta-explorer.energyweb.org/graphiql",
        EthereumNetwork.POLIS_MAINNET: "https://explorer.polis.tech/graphiql",
        EthereumNetwork.BOBA_NETWORK: "https://blockexplorer.boba.network/graphiql",
        EthereumNetwork.GATHER_DEVNET_NETWORK: "https://devnet-explorer.gather.network/graphiql",
        EthereumNetwork.GATHER_TESTNET_NETWORK: "https://testnet-explorer.gather.network/graphiql",
        EthereumNetwork.GATHER_MAINNET_NETWORK: "https://explorer.gather.network/graphiql",
        EthereumNetwork.METIS_GOERLI_TESTNET: "https://goerli.explorer.metisdevops.link/graphiql",
        EthereumNetwork.METIS_ANDROMEDA_MAINNET: "https://andromeda-explorer.metis.io/graphiql",
        EthereumNetwork.FUSE_MAINNET: "https://explorer.fuse.io/graphiql",
        EthereumNetwork.VELAS_EVM_MAINNET: "https://evmexplorer.velas.com/graphiql",
        EthereumNetwork.REI_NETWORK: "https://scan.rei.network/graphiql",
        EthereumNetwork.REI_CHAIN_TESTNET: "https://scan-test.rei.network/graphiql",
        EthereumNetwork.METER_MAINNET: "https://scan.meter.io/graphiql",
        EthereumNetwork.METER_TESTNET: "https://scan-warringstakes.meter.io/graphiql",
        EthereumNetwork.GODWOKEN_TESTNET_V1: "https://v1.betanet.gwscan.com/graphiql",
        EthereumNetwork.GODWOKEN_MAINNET: "https://v1.gwscan.com/graphiql",
        EthereumNetwork.VENIDIUM_TESTNET: "https://evm-testnet.venidiumexplorer.com/graphiql",
        EthereumNetwork.VENIDIUM_MAINNET: "https://evm.venidiumexplorer.com/graphiql",
        EthereumNetwork.KLAYTN_TESTNET_BAOBAB: "https://baobab.scope.klaytn.com/graphiql",
        EthereumNetwork.KLAYTN_MAINNET_CYPRESS: "https://scope.klaytn.com/graphiql",
        EthereumNetwork.ACALA_NETWORK: "https://blockscout.acala.network/graphiql",
        EthereumNetwork.KARURA_NETWORK_TESTNET: "https://blockscout.karura.network/graphiql",
        EthereumNetwork.ASTAR: "https://blockscout.com/astar/graphiql",
        EthereumNetwork.SHIDEN: "https://blockscout.com/shiden/graphiql",
        EthereumNetwork.EVMOS: "https://evm.evmos.org/graphiql",
        EthereumNetwork.EVMOS_TESTNET: "https://evm.evmos.dev/graphiql",
        EthereumNetwork.KCC_MAINNET: "https://scan.kcc.io/graphiql",
        EthereumNetwork.KCC_TESTNET: "https://scan-testnet.kcc.network/graphiql",
        EthereumNetwork.CROSSBELL: "https://scan.crossbell.io/graphiql",
        EthereumNetwork.ETHEREUM_CLASSIC: "https://blockscout.com/etc/mainnet/graphiql",
        EthereumNetwork.MORDOR_TESTNET: "https://blockscout.com/etc/mordor/graphiql",
        EthereumNetwork.SCROLL_SEPOLIA_TESTNET: "https://sepolia-blockscout.scroll.io/graphiql",
        EthereumNetwork.MANTLE: "https://explorer.mantle.xyz/api/v1/graphql",
        EthereumNetwork.MANTLE_TESTNET: "https://explorer.testnet.mantle.xyz/graphiql",
        EthereumNetwork.JAPAN_OPEN_CHAIN_MAINNET: "https://mainnet.japanopenchain.org/graphiql",
        EthereumNetwork.JAPAN_OPEN_CHAIN_TESTNET: "https://explorer.testnet.japanopenchain.org/graphiql",
        EthereumNetwork.ZETACHAIN_ATHENS_3_TESTNET: "https://zetachain-athens-3.blockscout.com/graphiql",
        EthereumNetwork.SCROLL: "https://blockscout.scroll.io/graphiql",
        EthereumNetwork.ROOTSTOCK_MAINNET: "https://rootstock.blockscout.com/graphiql",
        EthereumNetwork.ROOTSTOCK_TESTNET: "https://rootstock-testnet.blockscout.com/graphiql",
        EthereumNetwork.LINEA: "https://explorer.linea.build/graphiql",
        EthereumNetwork.LINEA_GOERLI: "https://explorer.goerli.linea.build/graphiql",
        EthereumNetwork.NEON_EVM_MAINNET: "https://neon.blockscout.com/graphiql",
        EthereumNetwork.NEON_EVM_DEVNET: "https://neon-devnet.blockscout.com/graphiql",
        EthereumNetwork.OASIS_SAPPHIRE: "https://explorer.sapphire.oasis.io/graphiql",
        EthereumNetwork.OASIS_SAPPHIRE_TESTNET: "https://testnet.explorer.sapphire.oasis.dev/graphiql",
        EthereumNetwork.CASCADIA_TESTNET: "https://explorer.cascadia.foundation/graphiql",
        EthereumNetwork.TENET: "https://tenetscan.io/graphiql",
        EthereumNetwork.TENET_TESTNET: "https://testnet.tenetscan.io/graphiql",
        EthereumNetwork.CRONOS_MAINNET: "https://cronos.org/explorer/graphiql",
        EthereumNetwork.CRONOS_TESTNET: "https://cronos.org/explorer/testnet3/graphiql",
        EthereumNetwork.THUNDERCORE_MAINNET: "https://explorer-mainnet.thundercore.com/graphiql",
        EthereumNetwork.THUNDERCORE_TESTNET: "https://explorer-testnet.thundercore.com/graphiql",
        EthereumNetwork.PGN_PUBLIC_GOODS_NETWORK: "https://explorer.publicgoods.network/graphiql",
        EthereumNetwork.SEPOLIA_PGN_PUBLIC_GOODS_NETWORK: "https://explorer.sepolia.publicgoods.network/graphiql",
        EthereumNetwork.ARTHERA_MAINNET: "https://explorer.arthera.net/graphiql",
        EthereumNetwork.ARTHERA_TESTNET: "https://explorer-test.arthera.net/graphiql",
        EthereumNetwork.MANTA_PACIFIC_MAINNET: "https://pacific-explorer.manta.network/graphiql",
        EthereumNetwork.KROMA: "https://blockscout.kroma.network/graphiql",
        EthereumNetwork.KROMA_SEPOLIA: "https://blockscout.sepolia.kroma.network/graphiql",
        EthereumNetwork.ZORA: "https://explorer.zora.energy/graphiql",
        EthereumNetwork.ZORA_SEPOLIA_TESTNET: "https://sepolia.explorer.zora.energy/graphiql",
        EthereumNetwork.HAQQ_NETWORK: "https://explorer.haqq.network/graphiql",
        EthereumNetwork.HAQQ_CHAIN_TESTNET: "https://explorer.testedge2.haqq.network/graphiql",
        EthereumNetwork.MODE: "https://explorer.mode.network/graphiql",
        EthereumNetwork.MODE_TESTNET: "https://sepolia.explorer.mode.network/graphiql",
        EthereumNetwork.MANTLE_SEPOLIA_TESTNET: "https://explorer.sepolia.mantle.xyz/api/v1/graphql",
        EthereumNetwork.OP_SEPOLIA_TESTNET: "https://optimism-sepolia.blockscout.com/graphiql",
        EthereumNetwork.UNREAL_TESTNET: "https://unreal.blockscout.com/graphiql",
        EthereumNetwork.TAIKO_KATLA_L2: "https://explorer.katla.taiko.xyz/graphiql",
        EthereumNetwork.SEI_DEVNET: "https://seitrace.com/graphiql",
        EthereumNetwork.LISK_SEPOLIA_TESTNET: "https://sepolia-blockscout.lisk.com/api/v1/graphql",
        EthereumNetwork.BOTANIX_TESTNET: "https://blockscout.botanixlabs.dev/graphiql",
        EthereumNetwork.REYA_NETWORK: "https://explorer.reya.network/graphiql",
        EthereumNetwork.AURORIA_TESTNET: "https://auroria.explorer.stratisevm.com/graphiql",
        EthereumNetwork.STRATIS_MAINNET: "https://explorer.stratisevm.com/graphiql",
        EthereumNetwork.SHIMMEREVM: "https://explorer.evm.shimmer.network/graphiql",
        EthereumNetwork.IOTA_EVM: "https://iota-evm.blockscout.com/graphiql",
        EthereumNetwork.BITROCK_MAINNET: "https://explorer.bit-rock.io/api/v1/graphql",
        EthereumNetwork.BITROCK_TESTNET: "https://testnetscan.bit-rock.io/api/v1/graphql",
        EthereumNetwork.OP_CELESTIA_RASPBERRY: "https://opcelestia-raspberry.gelatoscout.com/api/v1/graphql",
        EthereumNetwork.POLYGON_BLACKBERRY: "https://polygon-blackberry.gelatoscout.com/api/v1/graphql",
        EthereumNetwork.ARBITRUM_BLUEBERRY: "https://arb-blueberry.gelatoscout.com/api/v1/graphql",
        EthereumNetwork.RSS3_VSL_SEPOLIA_TESTNET: "https://scan.testnet.rss3.io/api/v1/graphql",
        EthereumNetwork.RSS3_VSL_MAINNET: "https://scan.rss3.io/api/v1/graphql",
        EthereumNetwork.CROSSFI_TESTNET: "https://scan.testnet.ms/graphiql",
        EthereumNetwork.ASTAR_ZKYOTO: "https://astar-zkyoto.blockscout.com/api/v1/graphql",
        EthereumNetwork.SAAKURU_MAINNET: "https://explorer.saakuru.network/graphiql",
        EthereumNetwork.REDSTONE: "https://explorer.redstone.xyz/api/v1/graphql",
        EthereumNetwork.GARNET_HOLESKY: "https://api.explorer.garnet.qry.live/api/v1/graphql",
        EthereumNetwork.TAIKO_HEKLA_L2: "https://blockscoutapi.hekla.taiko.xyz/graphiql",
        EthereumNetwork.ASTAR_ZKEVM: "https://astar-zkevm.explorer.startale.com/api/v1/graphql",
        EthereumNetwork.RE_AL: "https://explorer.re.al/api/v1/graphql",
        EthereumNetwork.UNREAL: "https://unreal.blockscout.com/api/v1/graphql",
        EthereumNetwork.LISK: "https://blockscout.lisk.com/api/v1/graphql",
        EthereumNetwork.OPEN_CAMPUS_CODEX: "https://opencampus-codex.blockscout.com/api/v1/graphql",
        EthereumNetwork.LORENZO: "https://scan.lorenzo-protocol.xyz/api/v1/graphql",
        EthereumNetwork.DODOCHAIN_TESTNET: "https://testnet-scan.dodochain.com/api/v1/graphql",
    }

    def __init__(self, network: EthereumNetwork):
        self.network = network
        self.grahpql_url = self.NETWORK_WITH_URL.get(network)
        if self.grahpql_url is None:
            raise BlockScoutConfigurationProblem(
                f"Network {network.name} - {network.value} not supported"
            )
        self.http_session = requests.Session()

    def build_url(self, path: str):
        return urljoin(self.grahpql_url, path)

    def _do_request(self, url: str, query: str) -> Optional[Dict[str, Any]]:
        response = self.http_session.post(url, json={"query": query}, timeout=10)
        if not response.ok:
            return None

        return response.json()

    def get_contract_metadata(
        self, address: ChecksumAddress
    ) -> Optional[ContractMetadata]:
        query = '{address(hash: "%s") { hash, smartContract {name, abi} }}' % address
        result = self._do_request(self.grahpql_url, query)
        if (
            result
            and "error" not in result
            and result.get("data", {}).get("address", {})
            and result["data"]["address"]["smartContract"]
        ):
            smart_contract = result["data"]["address"]["smartContract"]
            return ContractMetadata(
                smart_contract["name"], json.loads(smart_contract["abi"]), False
            )



# flake8: noqa F401
"""
Safe Addresses. Should be the same for every chain except for the ones with `chainId` protection. Check:
https://github.com/safe-global/safe-deployments/tree/main/src/assets

Safe V1.4.1: 0x41675C099F32341bf84BFc5382aF534df5C7461a
GnosisSafe V1.3.0: 0xd9Db270c1B5E3Bd161E8c8503c55cEABeE709552
GnosisSafe V1.1.1: 0x34CfAC646f301356fAa8B21e94227e3583Fe3F5F
GnosisSafe V1.1.0: 0xaE32496491b53841efb51829d6f886387708F99B
GnosisSafe V1.0.0: 0xb6029EA3B2c51D09a50B53CA8012FeEB05bDa35A

Factories
SafeProxyFactory V1.4.1: 0x4e1DCf7AD4e460CfD30791CCC4F9c8a4f820ec67
ProxyFactory V1.3.0: 0xa6B71E26C5e0845f74c812102Ca7114b6a896AB2
ProxyFactory V1.1.0: 0x50e55Af101C777bA7A1d560a774A82eF002ced9F
ProxyFactory V1.0.0: 0x12302fE9c02ff50939BaAaaf415fc226C078613C

FallbackHandler
CompatibilityFallBackHandler V1.4.1: 0xfd0732Dc9E303f09fCEf3a7388Ad10A83459Ec99
CompatibilityFallBackHandler V1.3.0: 0xf48f2B2d2a534e402487b3ee7C18c33Aec0Fe5e4

Libraries
CreateAndAddModules: 0x1a56aE690ab0818aF5cA349b7D21f1d7e76a3d36
MultiSend: 0x38869bf66a61cF6bDB996A6aE40D5853Fd43B526
MultiSendCallOnly: 0x9641d764fc13c8B624c04430C7356C1C7C8102e2
SimulateTxAccessor: 0x3d4BA2E0884aa488718476ca2FB8Efc291A46199
SignMessageLib: 0xd53cd0aB83D845Ac265BE939c57F53AD838012c9
"""

import json
import os
import sys
from functools import cache
from typing import Any, Callable, Dict, Optional

from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract

from .abis.multicall import multicall_v3_abi, multicall_v3_bytecode
from .contract_base import ContractBase

current_module = sys.modules[__name__]
contracts = {
    "compatibility_fallback_handler_V1_3_0": "CompatibilityFallbackHandler_V1_3_0.json",
    "compatibility_fallback_handler_V1_4_1": "CompatibilityFallbackHandler_V1_4_1.json",
    "cpk_factory": "CPKFactory.json",
    "delegate_constructor_proxy": "DelegateConstructorProxy.json",
    "erc1155": "ERC1155.json",
    "erc20": "ERC20.json",
    "erc721": "ERC721.json",
    "example_erc20": "ERC20TestToken.json",
    "kyber_network_proxy": "kyber_network_proxy.json",
    "multi_send": "MultiSend.json",
    "paying_proxy": "PayingProxy.json",
    "proxy": "Proxy_V1_1_1.json",
    "proxy_factory_V1_0_0": "ProxyFactory_V1_0_0.json",
    "proxy_factory_V1_1_1": "ProxyFactory_V1_1_1.json",
    "proxy_factory_V1_3_0": "ProxyFactory_V1_3_0.json",
    "proxy_factory_V1_4_1": "ProxyFactory_V1_4_1.json",
    "safe_V0_0_1": "GnosisSafe_V0_0_1.json",
    "safe_V1_0_0": "GnosisSafe_V1_0_0.json",
    "safe_V1_1_1": "GnosisSafe_V1_1_1.json",
    "safe_V1_3_0": "GnosisSafe_V1_3_0.json",
    "safe_V1_4_1": "Safe_V1_4_1.json",
    "simulate_tx_accessor_V1_4_1": "SimulateTxAccessor_V1_4_1.json",
    "sign_message_lib": "SignMessageLib.json",
    "uniswap_exchange": "uniswap_exchange.json",
    "uniswap_factory": "uniswap_factory.json",
    "uniswap_v2_factory": "uniswap_v2_factory.json",
    "uniswap_v2_pair": "uniswap_v2_pair.json",
    "uniswap_v2_router": "uniswap_v2_router.json",  # Router02
}


def load_contract_interface(file_name: str) -> Dict[str, Any]:
    """
    :param file_name:
    :return: Get parsed JSON to ABI with the relative filename to this file path
    """
    return _load_json_file(_abi_file_path(file_name))


def _abi_file_path(file_name: str) -> str:
    """
    :param file_name:
    :return: Full path to the provided ``file_name``
    """
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "abis", file_name))


def _load_json_file(path) -> Dict[str, Any]:
    """
    :param path:
    :return: Parsed json for the provided file
    """
    with open(path) as f:
        return json.load(f)


def generate_contract_fn(
    contract: Dict[str, Any]
) -> Callable[[Web3, Optional[ChecksumAddress]], Contract]:
    """
    Dynamically generate a function to build a Web3 Contract for the provided contract ABI

    :param contract:
    :return: function that will return a Web3 Contract from an ABI
    """

    def fn(w3: Web3, address: Optional[ChecksumAddress] = None) -> Contract:
        return w3.eth.contract(
            address=address, abi=contract["abi"], bytecode=contract.get("bytecode")
        )

    return fn


# Anotate functions that will be generated later with `setattr` so typing does not complain
def get_safe_contract(w3: Web3, address: Optional[ChecksumAddress] = None) -> Contract:
    """
    :param w3:
    :param address:
    :return: Latest available Safe Contract
    """
    return get_safe_V1_4_1_contract(w3, address=address)


def get_safe_V0_0_1_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_safe_V1_0_0_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_safe_V1_1_1_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_safe_V1_3_0_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_safe_V1_4_1_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_compatibility_fallback_handler_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    """
    :param w3:
    :param address: Usually a Safe address
    :return: Latest available Compatibility Fallback handler contract
    """
    return get_compatibility_fallback_handler_V1_4_1_contract(w3, address=address)


def get_compatibility_fallback_handler_V1_3_0_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_compatibility_fallback_handler_V1_4_1_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_sign_message_lib_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_erc20_contract(w3: Web3, address: Optional[ChecksumAddress] = None) -> Contract:
    pass


def get_erc721_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_erc1155_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_example_erc20_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_delegate_constructor_proxy_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_multi_send_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_paying_proxy_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_proxy_factory_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    """
    :param w3:
    :param address:
    :return: Latest available Safe Proxy Factory
    """
    return get_proxy_factory_V1_4_1_contract(w3, address=address)


def get_proxy_factory_V1_0_0_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_proxy_factory_V1_1_1_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_proxy_factory_V1_3_0_contract(
    w3: Web3, address: Optional[str] = None
) -> Contract:
    pass


def get_proxy_factory_V1_4_1_contract(
    w3: Web3, address: Optional[str] = None
) -> Contract:
    pass


def get_proxy_contract(w3: Web3, address: Optional[ChecksumAddress] = None) -> Contract:
    pass


def get_simulate_tx_accessor_V1_4_1_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_uniswap_exchange_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_uniswap_factory_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_uniswap_v2_factory_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_uniswap_v2_pair_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_uniswap_v2_router_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_kyber_network_proxy_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_cpk_factory_contract(
    w3: Web3, address: Optional[ChecksumAddress] = None
) -> Contract:
    pass


def get_multicall_v3_contract(w3: Web3, address: Optional[ChecksumAddress] = None):
    return w3.eth.contract(
        address=address,
        abi=multicall_v3_abi,
        bytecode=multicall_v3_bytecode,
    )


@cache
def get_proxy_1_0_0_deployed_bytecode() -> bytes:
    return HexBytes(load_contract_interface("Proxy_V1_0_0.json")["deployedBytecode"])


@cache
def get_proxy_1_1_1_deployed_bytecode() -> bytes:
    return HexBytes(load_contract_interface("Proxy_V1_1_1.json")["deployedBytecode"])


def get_proxy_1_1_1_mainnet_deployed_bytecode() -> bytes:
    """
    Somehow it's different from the generated version compiling the contracts
    """
    return HexBytes(
        "0x608060405273ffffffffffffffffffffffffffffffffffffffff600054167fa619486e0000000000000000000000000000000000000000000000000000000060003514156050578060005260206000f35b3660008037600080366000845af43d6000803e60008114156070573d6000fd5b3d6000f3fea265627a7a72315820d8a00dc4fe6bf675a9d7416fc2d00bb3433362aa8186b750f76c4027269667ff64736f6c634300050e0032"
    )


@cache
def get_proxy_1_3_0_deployed_bytecode() -> bytes:
    return HexBytes(load_contract_interface("Proxy_V1_3_0.json")["deployedBytecode"])


@cache
def get_proxy_1_4_1_deployed_bytecode() -> bytes:
    return HexBytes(load_contract_interface("Proxy_V1_4_1.json")["deployedBytecode"])


@cache
def get_paying_proxy_deployed_bytecode() -> bytes:
    return HexBytes(load_contract_interface("PayingProxy.json")["deployedBytecode"])


# Dynamically create the functions for getting the contracts
for contract_name, json_contract_filename in contracts.items():
    fn_name = "get_{}_contract".format(contract_name)
    contract_dict = load_contract_interface(json_contract_filename)
    if not contract_dict:
        raise ValueError(f"{contract_name} json cannot be empty")
    setattr(current_module, fn_name, generate_contract_fn(contract_dict))



from django.forms import forms
from django.test import TestCase

from hexbytes import HexBytes

from ...utils import fast_keccak_text
from ..forms import EthereumAddressFieldForm, HexFieldForm, Keccak256FieldForm


class EthereumAddressForm(forms.Form):
    value = EthereumAddressFieldForm()


class HexForm(forms.Form):
    value = HexFieldForm(required=False)


class Keccak256Form(forms.Form):
    value = Keccak256FieldForm(required=False)


class TestForms(TestCase):
    def test_ethereum_address_field_form(self):
        form = EthereumAddressForm(data={"value": "not an ethereum address"})
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["value"], ["Enter a valid checksummed Ethereum Address."]
        )

        form = EthereumAddressForm(
            data={"value": "0xbaa7df320f385318fe3409cc95db48de60dfa033"}
        )
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["value"], ["Enter a valid checksummed Ethereum Address."]
        )

        form = EthereumAddressForm(
            data={"value": "0xbaa7df320f385318fE3409CC95Db48DE60dfA033"}
        )
        self.assertTrue(form.is_valid())

    def test_hex_field_form(self):
        form = HexForm(data={"value": "not a hexadecimal"})
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["value"], ["Enter a valid hexadecimal."])

        form = HexForm(data={"value": "0xabcd"})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["value"], HexBytes("0xabcd"))

        form = HexForm(initial={"value": memoryview(bytes.fromhex("cdef"))})
        self.assertIn('value="0xcdef"', form.as_p())

        form = HexForm(data={"value": 1})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["value"], HexBytes(1))

        form = HexForm(data={"value": ""})
        self.assertTrue(form.is_valid())
        self.assertIsNone(form.cleaned_data["value"])

        form = HexForm(data={"value": None})
        self.assertTrue(form.is_valid())
        self.assertIsNone(form.cleaned_data["value"])

    def test_keccak256_field_form(self):
        form = Keccak256Form(data={"value": "not a hash"})
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["value"], ['"not a hash" is not a valid keccak256 hash.']
        )

        form = Keccak256Form(data={"value": "0x1234"})
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["value"], ['"0x1234" keccak256 hash should be 32 bytes.']
        )

        form = Keccak256Form(data={"value": fast_keccak_text("testing").hex()})
        self.assertTrue(form.is_valid())

        form = Keccak256Form(data={"value": ""})
        self.assertTrue(form.is_valid())
        self.assertIsNone(form.cleaned_data["value"])

        form = HexForm(data={"value": None})
        self.assertTrue(form.is_valid())
        self.assertIsNone(form.cleaned_data["value"])



from django.test import TestCase

import pytest

from ... import EthereumNetwork
from ...clients import EtherscanClient, EtherscanRateLimitError
from .mocks import etherscan_multisend_abi, sourcify_safe_metadata


class TestEtherscanClient(TestCase):
    def test_etherscan_get_abi(self):
        try:
            etherscan_api = EtherscanClient(EthereumNetwork.MAINNET)
            safe_master_copy_abi = sourcify_safe_metadata["output"]["abi"]
            safe_master_copy_address = "0x6851D6fDFAfD08c0295C392436245E5bc78B0185"
            self.assertEqual(
                etherscan_api.get_contract_abi(safe_master_copy_address),
                safe_master_copy_abi,
            )

            contract_metadata = etherscan_api.get_contract_metadata(
                safe_master_copy_address
            )
            self.assertEqual(contract_metadata.name, "GnosisSafe")
            self.assertEqual(contract_metadata.abi, safe_master_copy_abi)

            random_address = "0xaE32496491b53841efb51829d6f886387708F99a"
            self.assertIsNone(etherscan_api.get_contract_abi(random_address))
            self.assertIsNone(etherscan_api.get_contract_metadata(random_address))
        except EtherscanRateLimitError:
            self.skipTest("Etherscan rate limit reached")

    @pytest.mark.xfail(reason="Test might fail due to third-party service issues")
    def test_etherscan_get_abi_zksync(self):
        multisend_address = "0x0dFcccB95225ffB03c6FBB2559B530C2B7C8A912"
        etherscan_api = EtherscanClient(EthereumNetwork(324))
        self.assertEqual(
            etherscan_api.get_contract_abi(multisend_address),
            etherscan_multisend_abi,
        )



# flake8: noqa F401
from .cowswap import CowswapOracle
from .exceptions import (
    CannotGetPriceFromOracle,
    InvalidPriceFromOracle,
    OracleException,
)
from .kyber import KyberOracle
from .oracles import (
    AaveOracle,
    BalancerOracle,
    ComposedPriceOracle,
    CreamOracle,
    CurveOracle,
    EnzymeOracle,
    MooniswapOracle,
    PoolTogetherOracle,
    PriceOracle,
    PricePoolOracle,
    UnderlyingToken,
    UniswapOracle,
    UniswapV2Oracle,
    YearnOracle,
    ZerionComposedOracle,
)
from .superfluid import SuperfluidOracle
from .sushiswap import SushiswapOracle
from .uniswap_v3 import UniswapV3Oracle

__all__ = [
    "AaveOracle",
    "BalancerOracle",
    "ComposedPriceOracle",
    "CowswapOracle",
    "CreamOracle",
    "CurveOracle",
    "EnzymeOracle",
    "KyberOracle",
    "MooniswapOracle",
    "PoolTogetherOracle",
    "PriceOracle",
    "PricePoolOracle",
    "SushiswapOracle",
    "UnderlyingToken",
    "UniswapOracle",
    "UniswapV2Oracle",
    "UniswapV3Oracle",
    "YearnOracle",
    "ZerionComposedOracle",
    "SuperfluidOracle",
]
